/*
This file is part of the Juju GUI, which lets users view and manage Juju
environments within a graphical interface (https://launchpad.net/juju-gui).
Copyright (C) 2016 Canonical Ltd.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License version 3, as published by
the Free Software Foundation.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
General Public License for more details.

You should have received a copy of the GNU Affero General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

'use strict';

var juju = { components: {} }; // eslint-disable-line no-unused-vars

chai.config.includeStack = true;
chai.config.truncateThreshold = 0;

describe('DeploymentCredential', function () {
  var acl, clouds, credentials, regions;

  beforeAll(function (done) {
    // By loading this file it adds the component to the juju components.
    YUI().use('deployment-credential', function () {
      done();
    });
  });

  beforeEach(function () {
    credentials = [{ path: 'owner/test-cred' }];
    regions = ['test-region'];
  });

  beforeEach(function () {
    acl = { isReadOnly: sinon.stub().returns(false) };
    clouds = {
      google: {
        id: 'google',
        showLogo: true,
        signupUrl: 'https://console.cloud.google.com/billing/freetrial',
        svgHeight: 33,
        svgWidth: 256,
        title: 'Google Compute Engine'
      },
      azure: {
        id: 'azure',
        showLogo: true,
        signupUrl: 'https://azure.microsoft.com/en-us/free/',
        svgHeight: 24,
        svgWidth: 204,
        title: 'Microsoft Azure'
      },
      aws: {
        id: 'aws',
        showLogo: true,
        signupUrl: 'https://portal.aws.amazon.com/gp/aws/developer/' + 'registration/index.html',
        svgHeight: 48,
        svgWidth: 120,
        title: 'Amazon Web Services'
      },
      local: {
        id: 'local',
        showLogo: false,
        title: 'Local'
      }
    };
  });

  it('can display a loader when loading regions and credentials', function () {
    var renderer = jsTestUtils.shallowRender(React.createElement(juju.components.DeploymentCredential, {
      acl: acl,
      addTemplate: sinon.stub(),
      cloud: 'azure',
      clouds: clouds,
      listRegions: sinon.stub(),
      listTemplates: sinon.stub(),
      setCredential: sinon.stub(),
      setRegion: sinon.stub(),
      setTemplate: sinon.stub(),
      users: {},
      validateForm: sinon.stub() }), true);
    var output = renderer.getRenderOutput();
    var expected = React.createElement(
      'div',
      null,
      React.createElement(
        'div',
        { className: 'deployment-credential__loading' },
        React.createElement(juju.components.Spinner, null)
      )
    );
    assert.deepEqual(output, expected);
  });

  it('can render with a cloud', function () {
    var addTemplate = sinon.stub();
    var setCredential = sinon.stub();
    var setRegion = sinon.stub();
    var setTemplate = sinon.stub();
    var users = {};
    var validateForm = sinon.stub();
    var renderer = jsTestUtils.shallowRender(React.createElement(juju.components.DeploymentCredential, {
      acl: acl,
      addTemplate: addTemplate,
      cloud: 'azure',
      clouds: clouds,
      listRegions: sinon.stub().callsArgWith(1, null, []),
      listTemplates: sinon.stub().callsArgWith(0, null, []),
      setCredential: setCredential,
      setRegion: setRegion,
      setTemplate: setTemplate,
      users: users,
      validateForm: validateForm }), true);
    var instance = renderer.getMountedInstance();
    var output = renderer.getRenderOutput();
    var expected = React.createElement(
      'div',
      null,
      React.createElement(
        juju.components.ExpandingRow,
        {
          classes: { 'twelve-col': true },
          clickable: false,
          expanded: true },
        undefined,
        React.createElement(juju.components.DeploymentCredentialAdd, {
          acl: acl,
          addTemplate: addTemplate,
          close: instance._toggleAdd,
          cloud: 'azure',
          clouds: clouds,
          regions: [],
          setCredential: setCredential,
          setRegion: setRegion,
          setTemplate: setTemplate,
          users: users,
          validateForm: validateForm })
      )
    );
    assert.deepEqual(output, expected);
  });

  it('can render without a cloud', function () {
    var addTemplate = sinon.stub();
    var setCredential = sinon.stub();
    var setRegion = sinon.stub();
    var setTemplate = sinon.stub();
    var users = {};
    var validateForm = sinon.stub();
    var renderer = jsTestUtils.shallowRender(React.createElement(juju.components.DeploymentCredential, {
      acl: acl,
      addTemplate: addTemplate,
      cloud: null,
      clouds: clouds,
      listRegions: sinon.stub().callsArgWith(1, null, []),
      listTemplates: sinon.stub().callsArgWith(0, null, []),
      setCredential: setCredential,
      setRegion: setRegion,
      setTemplate: setTemplate,
      users: users,
      validateForm: validateForm }), true);
    var instance = renderer.getMountedInstance();
    var output = renderer.getRenderOutput();
    var expected = React.createElement(
      'div',
      null,
      React.createElement(
        juju.components.ExpandingRow,
        {
          classes: { 'twelve-col': true },
          clickable: false,
          expanded: true },
        undefined,
        React.createElement(juju.components.DeploymentCredentialAdd, {
          acl: acl,
          addTemplate: addTemplate,
          close: instance._toggleAdd,
          cloud: null,
          clouds: clouds,
          regions: [],
          setCredential: setCredential,
          setRegion: setRegion,
          setTemplate: setTemplate,
          users: users,
          validateForm: validateForm })
      )
    );
    assert.deepEqual(output, expected);
  });

  it('can show existing credentials', function () {
    var setCredential = sinon.stub();
    var setRegion = sinon.stub();
    var renderer = jsTestUtils.shallowRender(React.createElement(juju.components.DeploymentCredential, {
      acl: acl,
      addTemplate: sinon.stub(),
      cloud: 'azure',
      clouds: clouds,
      listRegions: sinon.stub().callsArgWith(1, null, regions),
      listTemplates: sinon.stub().callsArgWith(0, null, credentials),
      setCredential: setCredential,
      setRegion: setRegion,
      setTemplate: sinon.stub(),
      users: {},
      validateForm: sinon.stub() }), true);
    var instance = renderer.getMountedInstance();
    var output = renderer.getRenderOutput();
    var expected = React.createElement(
      'div',
      null,
      React.createElement(
        juju.components.ExpandingRow,
        {
          classes: { 'twelve-col': true },
          clickable: false,
          expanded: false },
        React.createElement(
          'form',
          { className: 'deployment-credential__form' },
          React.createElement(
            'div',
            { className: 'prepend-two four-col' },
            React.createElement(juju.components.InsetSelect, {
              disabled: false,
              label: 'Credential',
              onChange: instance._handleCredentialChange,
              options: [{
                label: 'owner/test-cred',
                value: 'owner/test-cred'
              }, {
                label: 'Add credential...',
                value: 'add-credential'
              }] })
          ),
          React.createElement(
            'div',
            { className: 'four-col' },
            React.createElement(juju.components.InsetSelect, {
              disabled: false,
              label: 'Region',
              onChange: setRegion,
              options: [{
                label: 'test-region',
                value: 'test-region'
              }] })
          )
        ),
        undefined
      )
    );
    assert.deepEqual(output, expected);
  });

  it('selects an initial credential', function () {
    var setCredential = sinon.stub();
    var setRegion = sinon.stub();
    jsTestUtils.shallowRender(React.createElement(juju.components.DeploymentCredential, {
      acl: acl,
      addTemplate: sinon.stub(),
      cloud: 'azure',
      clouds: clouds,
      listRegions: sinon.stub().callsArgWith(1, null, regions),
      listTemplates: sinon.stub().callsArgWith(0, null, credentials),
      setCredential: setCredential,
      setRegion: setRegion,
      setTemplate: sinon.stub(),
      users: {},
      validateForm: sinon.stub() }), true);
    assert.equal(setCredential.callCount, 1);
    assert.equal(setCredential.args[0][0], 'owner/test-cred');
  });

  it('can disable controls when read only', function () {
    acl.isReadOnly = sinon.stub().returns(true);
    var setCredential = sinon.stub();
    var setRegion = sinon.stub();
    var renderer = jsTestUtils.shallowRender(React.createElement(juju.components.DeploymentCredential, {
      acl: acl,
      addTemplate: sinon.stub(),
      cloud: 'azure',
      clouds: clouds,
      listRegions: sinon.stub().callsArgWith(1, null, regions),
      listTemplates: sinon.stub().callsArgWith(0, null, credentials),
      setCredential: setCredential,
      setRegion: setRegion,
      setTemplate: sinon.stub(),
      users: {},
      validateForm: sinon.stub() }), true);
    var instance = renderer.getMountedInstance();
    var output = renderer.getRenderOutput();
    var expected = React.createElement(
      'div',
      null,
      React.createElement(
        juju.components.ExpandingRow,
        {
          classes: { 'twelve-col': true },
          clickable: false,
          expanded: false },
        React.createElement(
          'form',
          { className: 'deployment-credential__form' },
          React.createElement(
            'div',
            { className: 'prepend-two four-col' },
            React.createElement(juju.components.InsetSelect, {
              disabled: true,
              label: 'Credential',
              onChange: instance._handleCredentialChange,
              options: [{
                label: 'owner/test-cred',
                value: 'owner/test-cred'
              }, {
                label: 'Add credential...',
                value: 'add-credential'
              }] })
          ),
          React.createElement(
            'div',
            { className: 'four-col' },
            React.createElement(juju.components.InsetSelect, {
              disabled: true,
              label: 'Region',
              onChange: setRegion,
              options: [{
                label: 'test-region',
                value: 'test-region'
              }] })
          )
        ),
        undefined
      )
    );
    assert.deepEqual(output, expected);
  });

  it('will abort the request when unmounting', function () {
    var abort = sinon.stub();
    var listTemplates = sinon.stub().returns({ abort: abort });
    var listRegions = sinon.stub().returns({ abort: abort });
    var renderer = jsTestUtils.shallowRender(React.createElement(juju.components.DeploymentCredential, {
      acl: acl,
      addTemplate: sinon.stub(),
      cloud: 'azure',
      clouds: clouds,
      listRegions: listRegions,
      listTemplates: listTemplates,
      setCredential: sinon.stub(),
      setRegion: sinon.stub(),
      setTemplate: sinon.stub(),
      users: {},
      validateForm: sinon.stub() }), true);
    renderer.unmount();
    assert.equal(abort.callCount, 2);
  });

  it('can navigate to the add credentials form', function () {
    var addTemplate = sinon.stub();
    var setCredential = sinon.stub();
    var setRegion = sinon.stub();
    var setTemplate = sinon.stub();
    var users = {};
    var validateForm = sinon.stub();
    var renderer = jsTestUtils.shallowRender(React.createElement(juju.components.DeploymentCredential, {
      acl: acl,
      addTemplate: addTemplate,
      cloud: 'azure',
      clouds: clouds,
      listRegions: sinon.stub().callsArgWith(1, null, regions),
      listTemplates: sinon.stub().callsArgWith(0, null, credentials),
      setCredential: setCredential,
      setRegion: setRegion,
      setTemplate: setTemplate,
      users: users,
      validateForm: validateForm }), true);
    var instance = renderer.getMountedInstance();
    instance._handleCredentialChange('add-credential');
    var output = renderer.getRenderOutput();
    // output = renderer.render();
    var expected = React.createElement(
      'div',
      null,
      React.createElement(
        juju.components.ExpandingRow,
        {
          classes: { 'twelve-col': true },
          clickable: false,
          expanded: true },
        undefined,
        React.createElement(juju.components.DeploymentCredentialAdd, {
          acl: acl,
          addTemplate: addTemplate,
          close: instance._toggleAdd,
          cloud: 'azure',
          clouds: clouds,
          regions: regions,
          setCredential: setCredential,
          setRegion: setRegion,
          setTemplate: setTemplate,
          users: users,
          validateForm: validateForm })
      )
    );
    assert.deepEqual(output, expected);
  });
});