/*
This file is part of the Juju GUI, which lets users view and manage Juju
environments within a graphical interface (https://launchpad.net/juju-gui).
Copyright (C) 2016 Canonical Ltd.

This program is free software: you can redistribute it and/or modify it under
the terms of the GNU Affero General Public License version 3, as published by
the Free Software Foundation.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero
General Public License for more details.

You should have received a copy of the GNU Affero General Public License along
with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

'use strict';

YUI.add('user-profile-model-list', function () {

  juju.components.UserProfileModelList = React.createClass({
    propTypes: {
      addNotification: React.PropTypes.func.isRequired,
      canCreateNew: React.PropTypes.bool.isRequired,
      currentModel: React.PropTypes.string,
      env: React.PropTypes.object.isRequired,
      hideConnectingMask: React.PropTypes.func.isRequired,
      jem: React.PropTypes.object,
      listModels: React.PropTypes.func.isRequired,
      showConnectingMask: React.PropTypes.func.isRequired,
      switchModel: React.PropTypes.func.isRequired,
      user: React.PropTypes.object,
      users: React.PropTypes.object.isRequired
    },

    getInitialState: function getInitialState() {
      this.xhrs = [];

      return {
        modelList: [],
        loadingModels: false,
        createNewModelActive: false
      };
    },

    componentWillMount: function componentWillMount() {
      this._fetchModels();
    },

    componentWillUnmount: function componentWillUnmount() {
      this.xhrs.forEach(function (xhr) {
        xhr && xhr.abort && xhr.abort();
      });
    },

    componentWillReceiveProps: function componentWillReceiveProps(nextProps) {
      var props = this.props;
      var currentUser = props.user && props.user.user;
      var nextUser = nextProps.user && nextProps.user.user;
      if (nextUser !== currentUser) {
        this._fetchModels();
      }
    },

    /**
      Makes a request of JEM or JES to fetch the user's availble models.
       @method _fetchModels
    */
    _fetchModels: function _fetchModels() {
      var _this = this;

      // Delay the call until after the state change to prevent race
      // conditions.
      this.setState({ loadingModels: true }, function () {
        var xhr = _this.props.listModels(_this._fetchModelsCallback);
        _this.xhrs.push(xhr);
      });
    },

    /**
      Callback for the JEM and JES list models call.
       @method _fetchModelsCallback
      @param {String} error The error from the request, or null.
      @param {Object} data The data from the request.
    */
    _fetchModelsCallback: function _fetchModelsCallback(error, data) {
      var _this2 = this;

      this.setState({ loadingModels: false }, function () {
        // We need to coerce error types returned by JES vs JEM into one error.
        var err = data.err || error;
        if (err) {
          console.error(err);
          return;
        }
        // data.models is only populated by Juju controllers, when using JEM
        // the models are in the top level 'data' object.
        var modelList;
        if (data.models) {
          modelList = data.models.map(function (model) {
            // XXX frankban: owner should be the ownerTag without the 'user-'
            // prefix here.
            model.owner = model.ownerTag;
            return model;
          });
        } else if (data.map) {
          modelList = data.map(function (model) {
            // XXX kadams54: JEM models don't *currently* have a name or owner.
            // They have a path which is a combination of both, but that format
            // may change on down the road. Hence this big comment.
            model.name = model.path;
            model.owner = model.path.split('/')[0];
            model.lastConnection = 'N/A';
            // XXX frankban: does JEM provide lifecycle indications?
            model.isAlive = true;
            return model;
          });
        }
        _this2.setState({ modelList: modelList });
      });
    },

    /**
       Take the supplied UUID, fetch the username and password then call the
      passed in switchModel method. If the UUID and name are null, simply
      disconnect without reconnecting to a new model.
       @method switchModel
      @param {String} uuid The model UUID.
      @param {String} name The model name.
      @param {Function} callback The function to be called once the model has
        been switched and logged into. Takes the following parameters:
        {Object} env The env that has been switched to.
    */
    switchModel: function switchModel(uuid, name, callback) {
      this.props.switchModel(uuid, this.state.modelList, name, callback);
    },

    /**
      Fetches the model name from the modelName ref and then creates a model
      then switches to it.
      @method createAndSwitch
      @param {Object} e The event handler from a form submission.
    */
    createAndSwitch: function createAndSwitch(e) {
      var _this3 = this;

      if (e && e.preventDefault) {
        e.preventDefault();
      }
      // The supplied new model name needs to be valid.
      var modelName = this.refs.modelName;
      if (!modelName.validate()) {
        // Only continue if the validation passes.
        return;
      }
      // Because this will automatically connect to the model lets show
      // the connecting mask right now.
      this.props.showConnectingMask();
      // XXX This is only for the JIMM flow.
      this.props.env.createModel(modelName.getValue(), this.props.user.user, function (data) {
        var err = data.err;
        if (err) {
          _this3.props.addNotification({
            title: 'Failed to create new Model',
            message: err,
            level: 'error'
          });
          console.error(err);
          _this3.props.hideConnectingMask(false);
          return;
        }
        _this3.switchModel(data.uuid, data.name);
      });
    },

    /**
      Generate the details for the provided model.
       @method _generateRow
      @param {Object} model A model object.
      @returns {Array} The markup for the row.
    */
    _generateRow: function _generateRow(model) {
      var uuid = model.uuid;
      var isCurrent = uuid === this.props.currentModel;
      if (!model.isAlive) {
        if (model.name) {
          return React.createElement(
            'li',
            { className: 'user-profile__entity user-profile__list-row',
              key: uuid },
            model.name,
            ' is being destroyed.'
          );
        } else {
          return null;
        }
      }
      return React.createElement(
        juju.components.UserProfileEntity,
        {
          entity: model,
          expanded: isCurrent,
          key: uuid,
          switchModel: this.switchModel,
          type: 'model' },
        React.createElement(
          'span',
          { className: 'user-profile__list-col three-col' },
          model.name || '--'
        ),
        React.createElement(
          'span',
          { className: 'user-profile__list-col four-col' },
          '--'
        ),
        React.createElement(
          'span',
          { className: 'user-profile__list-col two-col' },
          model.lastConnection || '--'
        ),
        React.createElement(
          'span',
          { className: 'user-profile__list-col one-col' },
          '--'
        ),
        React.createElement(
          'span',
          { className: 'user-profile__list-col two-col last-col' },
          model.owner || '--'
        )
      );
    },

    /**
      Generate the header for the models.
       @method _generateHeader
      @returns {Array} The markup for the header.
    */
    _generateHeader: function _generateHeader() {
      return React.createElement(
        'li',
        { className: 'user-profile__list-header twelve-col' },
        React.createElement(
          'span',
          { className: 'user-profile__list-col three-col' },
          'Name'
        ),
        React.createElement(
          'span',
          { className: 'user-profile__list-col four-col' },
          'Credential'
        ),
        React.createElement(
          'span',
          { className: 'user-profile__list-col two-col' },
          'Last accessed'
        ),
        React.createElement(
          'span',
          { className: 'user-profile__list-col one-col' },
          'Units'
        ),
        React.createElement(
          'span',
          { className: 'user-profile__list-col two-col last-col' },
          'Owner'
        )
      );
    },

    /**
      Depending on the existance of jem this will either switch to a
      disconnected model or open up the UI to allow the user to create a
      new model.
      @method _nextCreateStep
    */
    _nextCreateStep: function _nextCreateStep() {
      var _this4 = this;

      if (this.props.jem) {
        // Switch to a disconnected model
        this.switchModel();
      } else {
        // Open up the UI to specify a model name for the Controller.
        this.setState({ createNewModelActive: true }, function () {
          var input = _this4.refs.modelName;
          if (input && input.refs.field) {
            input.refs.field.focus();
          }
        });
      }
    },

    /**
      Generates the elements required for the create new button
      @method _generateCreateNew
      @param {String} className The class you'd like to have applied to the
        container.
    */
    _generateCreateNew: function _generateCreateNew(className) {
      var classes = classNames('user-profile__create-new', className, {
        collapsed: !this.state.createNewModelActive
      });
      return React.createElement(
        'div',
        { className: classes },
        React.createElement(
          'form',
          { onSubmit: this.createAndSwitch },
          React.createElement(juju.components.GenericButton, {
            action: this._nextCreateStep,
            type: 'inline-neutral first',
            title: 'Create new' }),
          React.createElement(juju.components.GenericInput, {
            placeholder: 'untitled_model',
            required: true,
            ref: 'modelName',
            validate: [{
              regex: /\S+/,
              error: 'This field is required.'
            }, {
              regex: /^[a-zA-Z0-9]([a-zA-Z0-9.-]*[a-zA-Z0-9])?$/,
              error: 'This field must only contain upper and lowercase ' + 'letters, numbers, and hyphens. It must not start or ' + 'end with a hyphen.'
            }] }),
          React.createElement(juju.components.GenericButton, {
            action: this.createAndSwitch,
            type: 'inline-neutral second',
            title: 'Submit' })
        )
      );
    },

    render: function render() {
      if (this.state.loadingModels) {
        return React.createElement(
          'div',
          { className: 'user-profile__model-list twelve-col' },
          React.createElement(juju.components.Spinner, null)
        );
      }
      var list = this.state.modelList;
      if (!list || list.length === 0) {
        return null;
      }
      var rows = list.map(this._generateRow);
      var createNewButton;
      if (this.props.canCreateNew) {
        createNewButton = this._generateCreateNew();
      }
      return React.createElement(
        'div',
        { className: 'user-profile__model-list' },
        React.createElement(
          'div',
          { className: 'user-profile__header twelve-col no-margin-bottom' },
          'Models',
          React.createElement(
            'span',
            { className: 'user-profile__size' },
            '(',
            list.length,
            ')'
          ),
          createNewButton
        ),
        React.createElement(
          'ul',
          { className: 'user-profile__list twelve-col' },
          this._generateHeader(),
          rows
        )
      );
    }

  });
}, '', {
  requires: ['generic-button', 'generic-input', 'loading-spinner', 'user-profile-entity']
});